==> linux-1.1.89.tar.gz: linux/fs/sysv/symlink.c <==
/*
 *  linux/fs/sysv/symlink.c
 *
 *  minix/symlink.c
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  coh/symlink.c
 *  Copyright (C) 1993  Pascal Haible, Bruno Haible
 *
 *  sysv/symlink.c
 *  Copyright (C) 1993  Bruno Haible
 *
 *  SystemV/Coherent symlink handling code
 */

#ifdef MODULE
#include <linux/module.h>
#endif

#include <asm/segment.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/sysv_fs.h>
#include <linux/stat.h>

static int sysv_readlink(struct inode *, char *, int);
static int sysv_follow_link(struct inode *, struct inode *, int, int, struct inode **);

/*
 * symlinks can't do much...
 */
struct inode_operations sysv_symlink_inode_operations = {
	NULL,			/* no file-operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	sysv_readlink,		/* readlink */
	sysv_follow_link,	/* follow_link */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL			/* permission */
};

static int sysv_follow_link(struct inode * dir, struct inode * inode,
	int flag, int mode, struct inode ** res_inode)
{
	int error;
	struct buffer_head * bh;

	*res_inode = NULL;
	if (!dir) {
		dir = current->fs->root;
		dir->i_count++;
	}
	if (!inode) {
		iput(dir);
		return -ENOENT;
	}
	if (!S_ISLNK(inode->i_mode)) {
		iput(dir);
		*res_inode = inode;
		return 0;
	}
	if (current->link_count > 5) {
		iput(inode);
		iput(dir);
		return -ELOOP;
	}
	if (!(bh = sysv_file_bread(inode, 0, 0))) { /* is reading 1 block enough ?? */
		iput(inode);
		iput(dir);
		return -EIO;
	}
	iput(inode);
	current->link_count++;
	error = open_namei(bh->b_data,flag,mode,res_inode,dir);
	current->link_count--;
	brelse(bh);
	return error;
}

static int sysv_readlink(struct inode * inode, char * buffer, int buflen)
{
	struct buffer_head * bh;
	char * bh_data;
	int i;
	char c;

	if (!S_ISLNK(inode->i_mode)) {
		iput(inode);
		return -EINVAL;
	}
	if (buflen > inode->i_sb->sv_block_size_1)
		buflen = inode->i_sb->sv_block_size_1;
	bh = sysv_file_bread(inode, 0, 0);
	iput(inode);
	if (!bh)
		return 0;
	bh_data = bh->b_data;
	i = 0;
	while (i<buflen && (c = bh_data[i])) {
		i++;
		put_fs_byte(c,buffer++);
	}
	brelse(bh);
	return i;
}

==> linux-2.3.30.tar.gz: linux/fs/sysv/symlink.c <==
/*
 *  linux/fs/sysv/symlink.c
 *
 *  minix/symlink.c
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  coh/symlink.c
 *  Copyright (C) 1993  Pascal Haible, Bruno Haible
 *
 *  sysv/symlink.c
 *  Copyright (C) 1993  Bruno Haible
 *
 *  SystemV/Coherent symlink handling code
 */

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/sysv_fs.h>
#include <linux/stat.h>

#include <asm/uaccess.h>

static int sysv_readlink(struct dentry *, char *, int);
static struct dentry *sysv_follow_link(struct dentry *, struct dentry *, unsigned int);

/*
 * symlinks can't do much...
 */
struct inode_operations sysv_symlink_inode_operations = {
	NULL,			/* no file-operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	sysv_readlink,		/* readlink */
	sysv_follow_link,	/* follow_link */
	NULL,			/* get_block */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL			/* revalidate */
};

static struct dentry *sysv_follow_link(struct dentry * dentry,
					struct dentry * base,
					unsigned int follow)
{
	struct inode *inode = dentry->d_inode;
	struct buffer_head * bh;

	bh = sysv_file_bread(inode, 0, 0);
	if (!bh) {
		dput(base);
		return ERR_PTR(-EIO);
	}
	UPDATE_ATIME(inode);
	base = lookup_dentry(bh->b_data, base, follow);
	brelse(bh);
	return base;
}

static int sysv_readlink(struct dentry * dentry, char * buffer, int buflen)
{
	struct inode *inode = dentry->d_inode;
	struct buffer_head * bh;
	char * bh_data;
	int i;
	char c;

	if (buflen > inode->i_sb->sv_block_size_1)
		buflen = inode->i_sb->sv_block_size_1;
	bh = sysv_file_bread(inode, 0, 0);
	if (!bh)
		return 0;
	bh_data = bh->b_data;
	i = 0;
	while (i<buflen && (c = bh_data[i])) {
		i++;
		put_user(c,buffer++);
	}
	brelse(bh);
	return i;
}
